
#include "Chip8VM.hpp"
#include "MainWindow.hpp"
#include "Utility.h"
#include <cstdarg>
#include <cassert>
#include <fstream>

//
//void MainWindowAppInterface::logMessage(const std::string& msg)
//{	
//	m_window->addLogMessage(msg);		
//}

const sf::Color KeypadDefaultBackground(135,175,135);
const sf::Color KeypadUseableBackground(105,175,105);
const sf::Color KeypadHoverColor(100, 100, 30);

const float toolbarHeight = 20;
const float debugFrameWidth = 200;
Chip8::Program program;

void MainWindowAppInterface::logMessage(const char* format, ...)
{
	//char* msg = new char[format.size()+15];
	char buffer[300];
	//const char* formatted = format.c_str();
	va_list arglist;
	va_start(arglist, format);
	vsnprintf(buffer, 300, format, arglist);
	//buffer[format.size()+15] = '\0';
	va_end(arglist);
	m_window->addLogMessage(buffer);

	//delete [] msg;
}

void MainWindowAppInterface::clearLog()
{
	m_window->clearLog();
}

/*###############################################################################################################################
MainWindow
*/

MainWindow::MainWindow() : m_chip8VM(nullptr), m_debugger(nullptr), m_showDebugFrame(true), m_showLogFrame(false)
{
	m_window.create(sf::VideoMode(1280, 720), "Chip8 Virtual Machine and debugger");

	m_gameScreenTexture.create(256, 192);
	m_gameScreenTexture.setRepeated(false);

	//setup GUI for the window
	m_mainVerticalPanel = std::make_shared<UI::Panel>(new UI::VerticalBoxLayout());
	m_mainVerticalPanel->setFillWidth(1);
	m_mainVerticalPanel->setBackgroundColor(sf::Color(0,0,0,0));

		createToolbarPanel();

		m_mainHorizontalPanel = std::make_shared<UI::Panel>(new UI::HorizontalBoxLayout());
		m_mainHorizontalPanel->setBackgroundColor(sf::Color(200,0,0,0));
		m_mainHorizontalPanel->setFillHeight(1);
		m_mainVerticalPanel->addEntity(m_mainHorizontalPanel);

			m_gameScreenRenderWidget = std::make_shared<UI::TextureRenderWidget>();
			m_gameScreenRenderWidget->setFillWidth(1);
			m_gameScreenRenderWidget->setFillHeight(1);
			m_gameScreenRenderWidget->setMargins(UI::Entity::Margins(5));
			m_gameScreenRenderWidget->setTexture(&m_gameScreenTexture.getTexture());
			m_mainHorizontalPanel->addEntity(m_gameScreenRenderWidget);

			createRightPanel();

	updateViews(1280, 720);
	
	sf::Event resizeEvent;
	resizeEvent.type = sf::Event::Resized;
	resizeEvent.size.width = 1280;
	resizeEvent.size.height = 720;

	m_mainVerticalPanel->onEvent(resizeEvent);

	m_window.setResizingCallback(std::bind(&MainWindow::onResizing, this, std::placeholders::_1, std::placeholders::_2));
	//const std::string fontFile = getExecutablePath() + "/resources/luximb.ttf";
	//m_font.loadFromFile(fontFile);


}

void MainWindow::createToolbarPanel()
{
	m_toolbarPanel = std::make_shared<UI::Panel>(new UI::HorizontalBoxLayout());
	m_toolbarPanel->setMinimumSize(sf::Vector2f(0,100));
	m_mainVerticalPanel->addEntity(m_toolbarPanel);

		UI::ButtonPtr buttonLoad(new UI::Button("Load program"));
		std::shared_ptr<sf::Texture> buttonLoadImage(new sf::Texture);
		buttonLoadImage->loadFromFile(getExecutablePath() + "/resources/MilkyLoad.png");
		buttonLoad->setImage(buttonLoadImage);
		buttonLoad->setMargins(UI::Button::Margins(5));;
		buttonLoad->setFillHeight(1);
		buttonLoad->setOnPressedCallback(std::bind(&MainWindow::onLoadProgram, this));
		m_toolbarPanel->addEntity(buttonLoad);

		UI::LabelPtr Separator(new UI::Label(""));
		Separator->setBorderStyle(UI::BorderStyle::Line);
		Separator->setDefaultSize(sf::Vector2f(1, 0));
		Separator->setBorderWidth(1);
		Separator->setBorderColor(UI::Default::TabHeader);
		Separator->setMargins(5,15,5,15);
		Separator->setFillHeight(1);
		m_toolbarPanel->addEntity(Separator);

		UI::ButtonPtr buttonStop(new UI::Button("Stop"));
		std::shared_ptr<sf::Texture> buttonStopImage(new sf::Texture);
		buttonStopImage->loadFromFile(getExecutablePath() + "/resources/MilkyStop.png");
		buttonStop->setImage(buttonStopImage);
		buttonStop->setMargins(UI::Button::Margins(5));
		buttonStop->setFillHeight(1);
		buttonStop->setOnPressedCallback([&] () {
			
		});
		m_toolbarPanel->addEntity(buttonStop);

		UI::ButtonPtr buttonPause(new UI::Button("Pause"));
		std::shared_ptr<sf::Texture> buttonPauseImage(new sf::Texture);
		buttonPauseImage->loadFromFile(getExecutablePath() + "/resources/MilkyPause.png");
		buttonPause->setImage(buttonPauseImage);
		buttonPause->setMargins(UI::Button::Margins(5));
		buttonPause->setFillHeight(1);
		buttonPause->setOnPressedCallback([&] () {
			
		});
		m_toolbarPanel->addEntity(buttonPause);

		UI::ButtonPtr buttonPlay(new UI::Button("Play"));
		std::shared_ptr<sf::Texture> buttonPlayImage(new sf::Texture);
		buttonPlayImage->loadFromFile(getExecutablePath() + "/resources/MilkyPlay.png");
		buttonPlay->setImage(buttonPlayImage);
		buttonPlay->setMargins(UI::Button::Margins(5));
		buttonPlay->setFillHeight(1);
		buttonPlay->setOnPressedCallback([&] () {
			
		});
		m_toolbarPanel->addEntity(buttonPlay);

		UI::ButtonPtr buttonStep(new UI::Button("Step"));
		std::shared_ptr<sf::Texture> buttonStepImage(new sf::Texture);
		buttonStepImage->loadFromFile(getExecutablePath() + "/resources/MilkyStepForward.png");
		buttonStep->setImage(buttonStepImage);
		buttonStep->setMargins(UI::Button::Margins(5));
		buttonStep->setFillHeight(1);
		buttonStep->setOnPressedCallback([&] () {
			
		});
		m_toolbarPanel->addEntity(buttonStep);

		UI::LabelPtr Separator2(new UI::Label(""));
		Separator2->setBorderStyle(UI::BorderStyle::Line);
		Separator2->setDefaultSize(sf::Vector2f(1, 0));
		Separator2->setBorderWidth(1);
		Separator2->setBorderColor(UI::Default::TabHeader);
		Separator2->setMargins(5,15,5,15);
		Separator2->setFillHeight(1);
		m_toolbarPanel->addEntity(Separator2);

		/*UI::PanelPtr toolbarFillerPanel(new UI::Panel(new UI::HorizontalBoxLayout()));
		toolbarFillerPanel->setFillWidth(1);
		m_toolbarPanel->addEntity(toolbarFillerPanel);*/

		UI::PanelPtr toolbarSavestatePanel(new UI::Panel(new UI::HorizontalBoxLayout()));
		//toolbarSavestatePanel->setFillWidth(1);
		toolbarSavestatePanel->setMinimumSize(sf::Vector2f(570,0));
		toolbarSavestatePanel->setBackgroundColor(sf::Color(0,0,0,0));
		m_toolbarPanel->addEntity(toolbarSavestatePanel);
			//add five save/load state buttons
			//create a black toolbar button sized image
			sf::RectangleShape blackRect(sf::Vector2f(96,48));
			blackRect.setPosition(0,0);
			blackRect.setFillColor(sf::Color::Black);

			for (int i = 0; i < 5; ++i)
			{
				UI::PanelPtr panel(new UI::Panel(new UI::VerticalBoxLayout()));
				panel->setMargins(5,5,0,5);
				panel->setBackgroundColor(sf::Color(80,80,80,200));
				//NOTE: the minimum size setting here should not be necessary, the information is already known by the system,
				//but I suppose we got a smallish bug there
				panel->setMinimumSize(sf::Vector2f(96,0));
				toolbarSavestatePanel->addEntity(panel);

				//see if we have saved states, if so we should assign the saved screenshots instead of an empty texture
				const std::string imagePath = getExecutablePath() + "/savestates/" + "screenshot" + toStr(i) + ".png";
				//const std::string statePath = getExecutablePath() + "/savestates/" + "savedstate" + toStr(i) +;
				sf::Texture screenshot;
				if (screenshot.loadFromFile(imagePath))
				{
					m_savestateScreen[i].create(96, 48);
					m_savestateScreen[i].draw(sf::Sprite(screenshot));
					m_savestateScreen[i].display();
				} else
				{
					m_savestateScreen[i].create(96,48);
					m_savestateScreen[i].draw(blackRect);
					m_savestateScreen[i].display();
				}
				//m_savestateScreen[i];
				UI::TextureRenderWidgetPtr trw(new UI::TextureRenderWidget(&m_savestateScreen[i].getTexture()));
				
				trw->setMargins(5,5,5,5);
				//trw->setMinimumSize(sf::Vector2f(48,24));
				panel->addEntity(trw);
				UI::PanelPtr buttonPanel(new UI::Panel(new UI::HorizontalBoxLayout()));
				buttonPanel->setBackgroundColor(sf::Color(0,0,0,0));
				panel->addEntity(buttonPanel);
				UI::ButtonPtr loadButton(new UI::Button("Load"));
				loadButton->setDefaultColor(UI::Default::PanelBackground);
				loadButton->setMargins(10,10,5,5);
				buttonPanel->addEntity(loadButton);
				loadButton->setOnPressedCallback([=] () {
					onLoadVMState(i);
				});
				UI::ButtonPtr saveButton(new UI::Button("Save"));
				saveButton->setDefaultColor(UI::Default::PanelBackground);
				saveButton->setMargins(14,10,5,5);
				buttonPanel->addEntity(saveButton);
				saveButton->setOnPressedCallback([=] () {
					onSaveVMState(i);
				});
			}
}

void MainWindow::createRightPanel()
{
	UI::PanelPtr rightPanel(new UI::Panel(new UI::VerticalBoxLayout()));
	rightPanel->setMinimumSize(sf::Vector2f(300,0));
	rightPanel->setMaximumSize(sf::Vector2f(300,0));

	m_mainHorizontalPanel->addEntity(rightPanel);

		m_debugPanelTab = std::make_shared<UI::TabbedInterface>();
		m_debugPanelTab->setFillHeight(1);
		m_debugPanelTab->setFillWidth(1);
		rightPanel->addEntity(m_debugPanelTab);

		UI::PanelPtr controlsPanel(new UI::Panel(new UI::VerticalBoxLayout()));
		//testPanel->setMinimumSize(sf::Vector2f(100,100));
		m_debugPanelTab->addPanel(controlsPanel, "Controls");
			UI::LabelPtr ControlsDesc(new UI::Label("The controls for Chip8/SuperChip use a 4x4 grid of keys, as the COSMAC VIP used to have. The default mapping is on the qwerty keyboard (1 to v)."));
			ControlsDesc->setMinimumSize(sf::Vector2f(0,70));
			ControlsDesc->setMargins(5,5,5,0);
			ControlsDesc->setCharacterSize(11);
			//ControlsDesc->setBorderStyle(UI::BorderStyle::Line);
			controlsPanel->addEntity(ControlsDesc);
			
			int i = 0;
			for (int y = 0; y < 4; ++y)
			{
				UI::PanelPtr KeypadRow(new UI::Panel(new UI::HorizontalBoxLayout()));
				KeypadRow->setMinimumSize(sf::Vector2f(20,20));
				KeypadRow->setFillWidth(1);
				
				for (int x = 0; x < 4; ++x)
				{
					m_keypadBackgrounds[i] = std::make_shared<UI::Panel>(new UI::HorizontalBoxLayout());
					m_keypadBackgrounds[i]->setMinimumSize(sf::Vector2f(0,40));
					m_keypadBackgrounds[i]->setFillWidth(1);
					//backgroundPanel->setFillHeight(1);
					m_keypadBackgrounds[i]->setBackgroundColor(KeypadDefaultBackground);
					m_keypadButtons[i] = std::make_shared<UI::Button>(KeyMappingDefaultStr[i]);
					//m_keypadButtons[i] = std::make_shared<UI::Button>(Chip8::KeyNames[i]);
					m_keypadButtons[i]->setMinimumSize(sf::Vector2f(20,20));
					m_keypadButtons[i]->setFillWidth(1);
					m_keypadButtons[i]->setFillHeight(1);
					m_keypadButtons[i]->setMargins(5,5,5,5);
					m_keypadButtons[i]->setHoverColor(KeypadHoverColor);
					//button->setMinimumSize(sf::Vector2f(20,20));
					m_keypadBackgrounds[i]->addEntity(m_keypadButtons[i]);
					KeypadRow->addEntity(m_keypadBackgrounds[i]);
					++i;
				}
				controlsPanel->addEntity(KeypadRow);
			}

		UI::PanelPtr vmPanel(new UI::Panel(new UI::VerticalBoxLayout()));
				
			UI::LabelPtr VMLogLabel(new UI::Label("VM log:"));
			VMLogLabel->setMargins(5,5,5,0);
			vmPanel->addEntity(VMLogLabel);

			UI::LabelPtr VMLogContent(new UI::Label([&]() -> std::string {
				/*std::string logOutput;
				for (int i = 0; i < m_logMessages.size(); ++i)
					logOutput += m_logMessages[i] + "\n";
				return logOutput;*/
				std::string output = "Mode: ";
				if (m_chip8VM != nullptr)
				{
					if (m_chip8VM->getCurrentState().superChipMode)
						output += "SuperChip8";
					else
						output += "Chip8";
					output += "\nStatus: ";
					output += Chip8::StatusNames[static_cast<int>(m_chip8VM->getExecuteStatus())];
					output += "\nSpeed: ";
					output += toStr(m_chip8VM->getCurrentState().frequency) + " Hz";
				} else
				{
					output +="\nStatus: ";
					output += "\nSpeed: ";
				}
				return output;
				//return "helloworldiamthemostamazingstringeverwritten.weshallconquertheworldwithmaplesyrup";
			}));
			VMLogContent->setMinimumSize(sf::Vector2f(0, 100));
			VMLogContent->setMargins(UI::Label::Margins(5));
			VMLogContent->setBorderStyle(UI::BorderStyle::Line);
			VMLogContent->setBorderColor(sf::Color::Transparent);
			vmPanel->addEntity(VMLogContent);

			UI::LabelPtr VMRegisterLabel(new UI::Label("Registers etc:"));
			VMRegisterLabel->setMargins(5,5,5,0);
			vmPanel->addEntity(VMRegisterLabel);

			UI::PanelPtr VMInfoDumpPanel(new UI::Panel(new UI::HorizontalBoxLayout()));
			VMInfoDumpPanel->setBackgroundColor(sf::Color::Transparent);
			VMInfoDumpPanel->setFillWidth(1);
			VMInfoDumpPanel->setFillHeight(1);
			vmPanel->addEntity(VMInfoDumpPanel);

			UI::PanelPtr VMRegisterDumpPanel(new UI::Panel(new UI::VerticalBoxLayout()));
			VMRegisterDumpPanel->setFillWidth(1);
			VMRegisterDumpPanel->setBackgroundColor(sf::Color::Transparent);
			VMInfoDumpPanel->addEntity(VMRegisterDumpPanel);
			for (int i = 0; i < 8; ++i)
			{
				UI::PanelPtr registerPanel(new UI::Panel(new UI::HorizontalBoxLayout()));
				registerPanel->setBackgroundColor(sf::Color::Transparent);
				registerPanel->setMinimumSize(sf::Vector2f(1,1));
				registerPanel->setMargins(20,5,20,5);
				registerPanel->setFillWidth(1);

				UI::LabelPtr registerALabel(new UI::Label(std::bind(&MainWindow::getChip8RegisterStr, this, i*2)));
				//registerALabel->setFillWidth(1);
				registerALabel->setMinimumSize(sf::Vector2f(60,20));
				//registerALabel->setMargins(20,5,20,0);
				registerALabel->setBorderStyle(UI::BorderStyle::Line);
				registerALabel->setBorderColor(sf::Color::Transparent);
				registerPanel->addEntity(registerALabel);

				UI::LabelPtr Separator(new UI::Label(""));
				Separator->setBorderStyle(UI::BorderStyle::Line);
				Separator->setDefaultSize(sf::Vector2f(1, 0));
				Separator->setBorderWidth(1);
				Separator->setBorderColor(UI::Default::TabHeader);
				Separator->setMargins(0,0,10,0);
				Separator->setFillHeight(1);
				registerPanel->addEntity(Separator);

				UI::LabelPtr registerBLabel(new UI::Label(std::bind(&MainWindow::getChip8RegisterStr, this, i*2+1)));
				//registerBLabel->setFillWidth(2);
				registerBLabel->setMinimumSize(sf::Vector2f(60,20));
				//registerBLabel->setMargins(20,5,20,0);
				registerBLabel->setBorderStyle(UI::BorderStyle::Line);
				registerBLabel->setBorderColor(sf::Color::Transparent);
				registerPanel->addEntity(registerBLabel);
				VMRegisterDumpPanel->addEntity(registerPanel);
			}

			UI::PanelPtr VMDataPanel(new UI::Panel(new UI::VerticalBoxLayout()));
			VMDataPanel->setFillWidth(1);
			VMDataPanel->setBackgroundColor(sf::Color::Transparent);
			//VMDataPanel->setMinimumSize(sf::Vector2f(1,1));
			VMInfoDumpPanel->addEntity(VMDataPanel);
			
			//TODO: fix VMDataPanel, it does not show up properly and it's children do not show at all

				UI::LabelPtr delayTimerLabel(new UI::Label([&] () -> std::string {
					if (m_chip8VM != nullptr)
						return "Delay timer:\n" + toStr(m_chip8VM->getCurrentState().timerDelay) + "(1/60th s)";
					else
						return "Delay timer: ";
				}));
				delayTimerLabel->setMargins(5,5,0,0);
				delayTimerLabel->setFillWidth(1);
				//delayTimerLabel->setBorderStyle(UI::BorderStyle::Line);
				VMDataPanel->addEntity(delayTimerLabel);
		m_debugPanelTab->addPanel(vmPanel, "VM");

		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		UI::PanelPtr debugPanel(new UI::Panel(new UI::VerticalBoxLayout()));
		//testPanel2->setMinimumSize(sf::Vector2f(100,100));
		m_debugPanelTab->addPanel(debugPanel, "Debug");

		UI::TabbedInterfacePtr debugTabTabbedInterface(new UI::TabbedInterface());
		debugTabTabbedInterface->setMargins(2,2,2,2);
		debugTabTabbedInterface->setFillHeight(1);
		debugTabTabbedInterface->setFillWidth(1);
		debugTabTabbedInterface->setBackgroundColor(UI::Default::TabHeader);
		
		debugTabTabbedInterface->setHeaderColor(UI::Default::TabHeaderBackground);

		
		debugPanel->addEntity(debugTabTabbedInterface);
			UI::PanelPtr disassemblyPanel(new UI::Panel(new UI::VerticalBoxLayout()));
			debugTabTabbedInterface->addPanel(disassemblyPanel, "Disassembly");
			UI::PanelPtr breakpointsPanel(new UI::Panel(new UI::VerticalBoxLayout()));
			debugTabTabbedInterface->addPanel(breakpointsPanel, "Breakpoints");

			debugTabTabbedInterface->selectTab(0);
			m_disassemblyListView = std::make_shared<UI::ListView>();
			m_disassemblyListView->setFillHeight(1);
			m_disassemblyListView->setMargins(5,5,5,5);
			//m_disassemblyListView->setItemCallback(&Debugger->DisassemblyItemsCallback);
			disassemblyPanel->addEntity(m_disassemblyListView);

	m_debugPanelTab->selectTab(0);
}

bool MainWindow::update()
{
	clearBuffer();

	sf::Event event;

	while (m_window.pollEvent(event))
	{
		m_mainVerticalPanel->onEvent(event);

		switch (event.type)
		{
			case sf::Event::Closed:
				m_window.close();
				return true;
				break;
			case sf::Event::Resized:
				//m_GUIView.reset(sf::FloatRect(0, 0, event.size.width, event.size.height));
				//updateViews(event.size.width, event.size.height);
				break;
			case sf::Event::KeyPressed:
				
				switch (event.key.code)
				{
					case sf::Keyboard::Escape:
						return true;
						break;
					case sf::Keyboard::O:
						if (event.key.control)
						{
							onLoadProgram();
						}
						break;
					case sf::Keyboard::F10:
						m_chip8VM->step(DesiredFrameTime);
						break;
					case sf::Keyboard::F4:
						break;
					case sf::Keyboard::F5:
						m_showDebugFrame = !m_showDebugFrame;
						break;
					case sf::Keyboard::F6:
						m_showLogFrame = !m_showLogFrame;
						break;
				}
				break;
			case sf::Event::MouseButtonPressed:
				break;
			case sf::Event::MouseButtonReleased:
				break;
			case sf::Event::MouseMoved:
				break;
		}
		
		
	}
	
	if (m_chip8VM != nullptr)
	{
		const bool* keypadActivity = m_chip8VM->getKeypadActivity();
		for (int i = 0; i < 16; ++i)
		{
			if (keypadActivity[i])
			{
				m_keypadBackgrounds[i]->setBackgroundColor(KeypadUseableBackground);
				//m_keypadButtons[i]->setMargins(5,5,5,5);
			} else
			{
				m_keypadBackgrounds[i]->setBackgroundColor(KeypadDefaultBackground);
				//m_keypadButtons[i]->setMargins(0,0,0,0);
			}
		}
	}

	drawEverything();
	m_window.display();

	return false;
}

void MainWindow::clearBuffer()
{
	m_window.clear();
}

/*
Draw the passed in Chip8 video buffer to the designated game view
We do this by setting a virtual coordinate system of the same size as the current Chip8 VM video size,
and then ensuring we keep the correct aspect ratio according to the current window size.
*/
void MainWindow::drawChip8Buffer(const uint8_t* buffer, unsigned int width, unsigned int height)
{

	sf::RectangleShape rect( sf::Vector2f(1,1) );

	m_gameView.reset( sf::FloatRect(0, 0, width, height) );

	//calculate this so we fit in the center, not overlapping with the GUI anywhere
	//we should also calculate a fitting size that preserves the correct ratio
	float x1, y1, x2, y2;

	sf::Vector2f windowSize = m_gameScreenRenderWidget->getSize();//m_window.getView().getSize();

	//sf::Vector2f toolbarsize = m_toolbarPanel->getSize();
	x1 = 0.f; 
	y1 = 0;
	
	if (windowSize.y > (windowSize.x / 2.f))
	{
		x2 = windowSize.x;
		y2 = (x2 / 2.f);
	} else
	{
		y2 = windowSize.y;
		x2 = y2 * 2.f;
		
	}

	x2 = x2 / windowSize.x;
	y2 = y2 / windowSize.y;
	


	m_gameView.setViewport( sf::FloatRect(x1, y1, x2, y2) );

	m_gameScreenTexture.setView(m_gameView);
	
	//draw a fullscreen rect behind to check the viewport
	sf::RectangleShape background( sf::Vector2f(width, height) );
	background.setPosition(0,0);
	background.setFillColor(sf::Color(0,200,0));
	m_gameScreenTexture.clear();
	m_gameScreenTexture.draw(background);

	for (int y = 0; y < height; ++y)
	{
		for (int x = 0; x < width; ++x)
		{
			rect.setPosition(x, y);
			uint8_t color = buffer[x + y*width];
			if (color > 0)
			{
				rect.setFillColor( sf::Color(color, color, color) );
				m_gameScreenTexture.draw(rect);
			}
		}
	}

	m_gameScreenTexture.display();
}

void MainWindow::addLogMessage(const std::string& msg)
{
	m_logMessages.push_back(msg);
}

void MainWindow::clearLog()
{
	m_logMessages.clear();
}

void MainWindow::insertVMPointers(Chip8::VirtualMachine* vm, Debugger* debugger)
{
	m_chip8VM = vm;
	m_debugger = debugger;
	if (debugger != nullptr)
		m_disassemblyListView->setItemCallback(&m_debugger->DisassemblyItemsCallback);
}

void MainWindow::onLoadProgram()
{
	std::string file = "";
	bool fileChosen = showFileOpenDialog(m_window.getSystemHandle(), file);
	//bool fileChosen = true;
	if (fileChosen)
	{
		//addLogMessage(file);
		//TODO: read from file
		//std::string file = "E:\\Development\\Projects\\Chip8Emu\\Chip8Emu\\resources\\SuperChip Games\\Ant - In Search of Coke [Erin S. Catto].ch8";
		std::string filecontents = readFile(file);
		program.bytecode.clear();
		program.bytecode.insert(program.bytecode.begin(), filecontents.c_str(), filecontents.c_str() + filecontents.size());
		m_chip8VM->load(&program);
								
	}
}

void MainWindow::onResizing(unsigned int w, unsigned int h)
{
	clearBuffer();
	updateViews(w, h);
	m_window.setView(m_GUIView);

	//adjust debugger output
	if (m_debugger)
	{
		int linesToShow = std::floor(m_disassemblyListView->getSize().y / static_cast<float>(UI::Default::TextFont->getLineSpacing(10)));
		m_debugger->showLines(linesToShow);
	}

	sf::Event event;
	event.type = sf::Event::Resized;
	event.size.width = w;
	event.size.height = h;
	m_mainVerticalPanel->onEvent(event);
	
	drawEverything();

	m_window.display();
}

void MainWindow::onLoadVMState(int index)
{
	//const std::string imagePath = getExecutablePath() + "/savestates/" + "screenshot" + toStr(index) + ".png";
	const std::string statePath = getExecutablePath() + "/savestates/" + "savedstate" + toStr(index) + ".c8state";

	std::ifstream fin(statePath, std::ifstream::binary);
	if (fin.is_open())
	{
		Chip8::VirtualMachine::State loadedState;
		fin.read(reinterpret_cast<char*>(&loadedState), sizeof loadedState);
		fin.close();
		m_chip8VM->loadState(loadedState);
	}
}

void MainWindow::onSaveVMState(int index)
{
	const std::string imagePath = getExecutablePath() + "/savestates/" + "screenshot" + toStr(index) + ".png";
	const std::string statePath = getExecutablePath() + "/savestates/" + "savedstate" + toStr(index) + ".c8state";

	sf::Sprite sprite(m_gameScreenTexture.getTexture());
	
	sf::Vector2u textureSize = m_gameScreenTexture.getTexture().getSize();
	sf::Vector2f screenWidget = m_gameScreenRenderWidget->getSize();
	sprite.setScale(96 / static_cast<float>(textureSize.x), 48 / static_cast<float>(textureSize.y));
	sf::View view;
	view.reset(sf::FloatRect(0,0,96,48));
	//reverse the distortion caused by the screen rendering in drawChip8Buffer()
	view.setViewport( sf::FloatRect(0, 0, (1 / m_gameScreenTexture.getView().getViewport().width), (1 / m_gameScreenTexture.getView().getViewport().height)) );

	m_savestateScreen[index].clear();
	m_savestateScreen[index].setView(view);
	m_savestateScreen[index].draw(sprite);
	m_savestateScreen[index].display();
	sf::Image img;
	
	img = m_savestateScreen[index].getTexture().copyToImage();
	img.saveToFile(imagePath);
	
	Chip8::VirtualMachine::State savedState = m_chip8VM->saveState();

	//we solve savestates with the naive non-portable version
	//this will probably work decently well everywhere, but moving them across platforms is not guaranteed to work, and neither is
	//it guaranteed to work if we were to change compiler settings or compiler
	std::ofstream fout(statePath, std::ofstream::binary | std::ofstream::trunc);
	fout.write(reinterpret_cast<const char*>(&savedState), sizeof savedState);
	fout.flush();
	fout.close();
}

std::string MainWindow::getChip8RegisterStr(int index)
{
	if (m_chip8VM != nullptr)
		return "V" + toStrHex(index) + ": " + toStrHex(m_chip8VM->getCurrentState().registerV[index]);
	else
		return "V" + toStrHex(index) + ": ";
}

void MainWindow::drawEverything()
{
	m_mainVerticalPanel->onDraw(&m_window);
	drawChip8Buffer(m_chip8VM->getHardwareOutput().videoBuffer, m_chip8VM->getHardwareOutput().width, m_chip8VM->getHardwareOutput().height);
}

void MainWindow::drawGUI()
{
	/*
	We need to use getView().getSize() rather than the getSize() from window directly, because SFML does not change the render viewport's coordinates
	when the OS window changes size.
	*/
	
	//const sf::Vector2f windowSize = m_window.getView().getSize();
	//
	//const float logFrameWidth = m_showDebugFrame ? windowSize.x - debugFrameWidth : windowSize.x;

	////Draw toolbar
	////TODO: use texture to provide better graphics
	//{
	//	sf::RectangleShape loadButton(sf::Vector2f(100,toolbarHeight));
	//	loadButton.setPosition(20, 0);
	//	m_window.draw(loadButton);

	//	sf::Text loadText("Load program", m_font, 11);
	//	loadText.setColor(sf::Color::Black);
	//	loadText.setPosition(25, 5);
	//	m_window.draw(loadText);
	//}
	//{
	//	sf::RectangleShape stepButton(sf::Vector2f(100,toolbarHeight));
	//	stepButton.setPosition(125, 0);
	//	m_window.draw(stepButton);

	//	sf::Text stepText("Step", m_font, 11);
	//	stepText.setColor(sf::Color::Black);
	//	stepText.setPosition(150, 5);
	//	m_window.draw(stepText);
	//}

	////End draw toolbar

	//if (m_showDebugFrame)
	//{
	//	sf::RectangleShape debugFrameCPU(sf::Vector2f(debugFrameWidth, windowSize.y - toolbarHeight));
	//	debugFrameCPU.setFillColor(sf::Color(90, 90, 90));
	//	debugFrameCPU.setPosition(windowSize.x - debugFrameWidth, toolbarHeight);
	//	m_window.draw(debugFrameCPU);

	//	const float debugTextPosition = windowSize.x - debugFrameWidth;
	//	sf::Text blackText;
	//	blackText.setFont(m_font);
	//	blackText.setCharacterSize(11);
	//	blackText.setColor(sf::Color::Black);

	//	switch (m_debugFrameTab)
	//	{
	//		case TAB_VMDATA:
	//		{
	//			for (int i = 0; i < m_logMessages.size(); ++i)
	//			{
	//				blackText.setString(m_logMessages[i]);
	//				blackText.setPosition(debugTextPosition + 5, toolbarHeight + 10 + (12*i));

	//				m_window.draw(blackText);
	//			}

	//			//std::list<Chip8::VirtualMachine::OpcodeEntry> execHistory = Chip8VM->getExecutionHistory();

	//			int i = 1;
	//			int usedHeight = toolbarHeight + 10 + ((m_logMessages.size()) * 12);
	//			//TODO: calculate how many we can fit on the screen and only show the last few
	//			/*int messagesToFit = ((windowSize.y - toolbarHeight) - usedHeight) / 12.f;

	//			for (auto iter = execHistory.begin(); iter != execHistory.end(); ++iter)
	//			{
	//				if (i > (execHistory.size() - messagesToFit))
	//				{
	//					blackText.setString(toStr(i) + " - 0x" + toStrHex(iter->pc) + ": 0x" + toStrHex(iter->opcode));
	//					blackText.setPosition(debugTextPosition + 5, usedHeight + (12*(i - (execHistory.size() - messagesToFit))));

	//					m_window.draw(blackText);
	//				}
	//				++i;
	//			}*/

	//			int messagesToFit = ((windowSize.y - toolbarHeight) - usedHeight) / 12.f;
	//			Debugger->showLines(messagesToFit - 1);
	//			if (Debugger)
	//			{
	//				for (auto iter = Debugger->getCurrentExecutionList().begin(); iter != Debugger->getCurrentExecutionList().end(); ++iter)
	//				{
	//					if (i > static_cast<signed int>(Debugger->getCurrentExecutionList().size() - messagesToFit))
	//					{
	//						blackText.setString(iter->string);
	//						if (iter->current)
	//						{
	//							sf::RectangleShape textBackground(sf::Vector2f(blackText.getGlobalBounds().width, 12));
	//							textBackground.setFillColor(sf::Color::Red);
	//							textBackground.setPosition(debugTextPosition + 5, usedHeight + (12*(i - (Debugger->getCurrentExecutionList().size() - messagesToFit))) );
	//							m_window.draw(textBackground);
	//						}
	//						blackText.setPosition(debugTextPosition + 5, usedHeight + (12*(i - (Debugger->getCurrentExecutionList().size() - messagesToFit))) );

	//						m_window.draw(blackText);
	//					}
	//					++i;
	//				}
	//			}

	//			break;
	//		}
	//		case TAB_KEYPAD:
	//		{
	//			const int TabKeypadMargin = 10;
	//			const bool* keypadActivity = Chip8VM->getKeypadActivity();

	//			const float startX = debugTextPosition + TabKeypadMargin;
	//			const float startY = toolbarHeight + TabKeypadMargin;

	//			const float endX = debugTextPosition + debugFrameWidth - TabKeypadMargin;
	//			const float endY = windowSize.y - TabKeypadMargin;

	//			const float keypadWidth = (endX - startX) / 4.f;
	//			const float keypadHeight = keypadWidth;

	//			/*sf::RectangleShape fillrect(sf::Vector2f(endX - startX, endY - startY));
	//			fillrect.setPosition(startX, startY);
	//			m_window.draw(fillrect);
	//			*/
	//			int i = 0;
	//			for (int y = 0; y < 4; ++y)
	//			{
	//				for (int x = 0; x < 4; ++x)
	//				{
	//					const float xpos = startX + (keypadWidth * x);
	//					const float ypos = startY + (keypadHeight * y);

	//					sf::RectangleShape rect(sf::Vector2f(keypadWidth, keypadHeight));
	//					if ((x+y) % 2 == 1)
	//						rect.setFillColor(sf::Color(170, 170, 170));
	//					else
	//						rect.setFillColor(sf::Color(150, 150, 150));
	//					blackText.setString(Chip8::KeyNames[i]);
	//					//the magic numbers 9 and 6 here are half the size of the characters in the string
	//					//they are used for centering the text
	//					blackText.setPosition( xpos + (keypadWidth / 2.f) - 9, ypos + (keypadHeight / 2.f) - 6);
	//					rect.setPosition( xpos, ypos);
	//					m_window.draw(rect);
	//					if (keypadActivity[i])
	//					{
	//						sf::RectangleShape activeKeyRect(sf::Vector2f(keypadWidth - 15, keypadHeight - 15));
	//						activeKeyRect.setPosition(xpos + 7.5f, ypos + 7.5f);
	//						if ((x+y) % 2 == 0)
	//							activeKeyRect.setFillColor(sf::Color(170, 170, 170));
	//						else
	//							activeKeyRect.setFillColor(sf::Color(150, 150, 150));
	//						m_window.draw(activeKeyRect);
	//					}
	//					m_window.draw(blackText);
	//					++i;
	//				}
	//			}
	//			break;
	//		}
	//	}
	//	

	//	//m_logMessages.clear();
	//}

	//if (m_showLogFrame)
	//{
	//	sf::RectangleShape logFrame(sf::Vector2f(logFrameWidth, windowSize.y * 0.2f));
	//	logFrame.setFillColor(sf::Color(170, 170, 170));
	//	logFrame.setPosition( 0, windowSize.y * 0.8f);
	//	m_window.draw(logFrame);
	//}

	
}

void MainWindow::updateViews(float w, float h)
{
	m_GUIView.reset(sf::FloatRect(0, 0, w, h));
	//m_gameView.reset(sf::FloatRect(0, 0, w, h));
	//m_GUIView.setViewport(sf::FloatRect(0, 0, 0.5, 0.5));
	//TODO: calculate proper size of GameView
	//m_gameView.setViewport(sf::FloatRect(0, 0, 1, 1));
}