/**
A slightly ugly part about this is the hardcoded GUI I've written. Ideally it should be moved out into separate classes.
But it still works fine for a small application with specific needs, and is faster to implement this way, avoiding the
need for generalisation.

*/
#pragma once

#include "Chip8VM.hpp"
#include "CustomSFMLGUI.h"

#include "SFML/Graphics.hpp"
#include <cstdint>
#include <memory>
#include <vector>
#include <string>

//the optimal time for a refresh rate of 60 FPS
const uint32_t DesiredFrameTime = 16667;


const sf::Keyboard::Key KeyMappingDefault[0x10] = {
	sf::Keyboard::Num1, sf::Keyboard::Num2, sf::Keyboard::Num3, sf::Keyboard::Num4,

	sf::Keyboard::Q, sf::Keyboard::W, sf::Keyboard::E, sf::Keyboard::R,

	sf::Keyboard::A, sf::Keyboard::S, sf::Keyboard::D, sf::Keyboard::F,

	sf::Keyboard::Z, sf::Keyboard::X, sf::Keyboard::C, sf::Keyboard::V
};

static const char* KeyMappingDefaultStr[0x10] = {
	"1", "2", "3", "4",
	"Q", "W", "E", "R",
	"A", "S", "D", "F",
	"Z", "X", "C", "V"
};

class MainWindowAppInterface : public Chip8::VirtualMachine::IApplicationFrontend
{
	public:
		MainWindowAppInterface(class MainWindow* w) : m_window(w)
		{}

		virtual bool getKeyDown(Chip8::Key key)
		{
			if (sf::Keyboard::isKeyPressed(KeyMappingDefault[static_cast<int>(key)]))
				return true;
			return false;
		}

		//virtual void logMessage(const std::string& msg);
		virtual void logMessage(const char* format, ...);
		virtual void clearLog();

	private:
		class MainWindow* m_window;
};





class MainWindow
{
	public:
		MainWindow();

		//return true when the window is closed (= exit program)
		bool update();

		void clearBuffer();

		void drawChip8Buffer(const uint8_t* buffer, unsigned int width, unsigned int height);

		void addLogMessage(const std::string& msg);
		void clearLog();

		/*enum DebugMsgType {
			DMT_OPLIST, DMT_CURRENTOP, DMT_REGISTERS
		};
		void addDebugMessage(int type, const std::string& msg);*/

		void insertVMPointers(Chip8::VirtualMachine*, class Debugger*);

	private:
		void onLoadProgram();

		void onResizing(unsigned int w, unsigned int h);
		void onLoadVMState(int);
		void onSaveVMState(int);

		std::string getChip8RegisterStr(int index);

		void drawEverything();
		void drawGUI();
		void updateViews(float w, float h);

		void createToolbarPanel();
		void createRightPanel();

		Chip8::VirtualMachine*	m_chip8VM;
		class Debugger*			m_debugger;
		
		sf::RenderWindow m_window;
		sf::View m_GUIView;
		sf::View m_gameView;
		//sf::Font m_font;

		//TODO: rename DebugFrame to something like InfoFrame
		bool m_showDebugFrame;
		bool m_showLogFrame;

		std::vector<std::string> m_logMessages;
		std::vector<std::string> m_debugMessages;

		//UI stuff
		UI::PanelPtr m_mainVerticalPanel;
		UI::PanelPtr m_mainHorizontalPanel;
						
		UI::PanelPtr m_gameScreenPanel;
		UI::PanelPtr m_toolbarPanel;
						
		//std::shared_ptr<UI::Panel> m_debugPanel;
		UI::TabbedInterfacePtr m_debugPanelTab;

		UI::TextureRenderWidgetPtr m_gameScreenRenderWidget;
		UI::ListViewPtr m_disassemblyListView;

		sf::RenderTexture m_gameScreenTexture;

		sf::RenderTexture m_savestateScreen[5];

		UI::PanelPtr m_keypadBackgrounds[16];
		UI::ButtonPtr m_keypadButtons[16];
};



class Debugger : public Chip8::IDebugger
{
	public:
		struct ExecutionEntry
		{
			bool current;
			std::string string;
		};

		class DisassemblyItems : public UI::ListView::IItemCallback
		{
			public:
				std::vector<ExecutionEntry>* ExecutionEntries;

				virtual int getNumItems();
				virtual sf::Vector2f getSize(int index);
				virtual void draw(int index, const sf::Vector2f& position, const sf::Vector2f& size, sf::RenderTarget*);
			private:

		};


		Debugger();

		virtual void onAttach(Chip8::VirtualMachine* );
		virtual bool onStep(uint16_t opcode);
		
		void resume();
		
		void showLines(int);

		void setBreakpoint(uint16_t position);
		void removeBreakpoint(uint16_t position);

		

		const std::vector<ExecutionEntry>& getCurrentExecutionList() { return m_execution; }

		//TODO: make a separate disassembler class?
		//this could support disassembling entire programs as well, but that can also be done by just looping over this function
		std::string disassemble(uint16_t opcode);

		DisassemblyItems DisassemblyItemsCallback;
	private:
		Chip8::VirtualMachine*	m_vm;

		int m_executionOutputLines;

		std::list<uint16_t> m_breakpoints;
		std::vector<ExecutionEntry> m_execution;
		std::string m_registerData;
};