#include "TestChip8VM.hpp"
#include <cassert>

#define CHIP8_TESTS_ASSERT

bool TestSuiteChip8VM::runAll()
{
	return false;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Chip8 tests

bool TestSuiteChip8VM::testClearDisplay()
{
	/*
	Test the clear screen instruction.
	If any byte of the video buffer is non-zero after this, it failed.
	For a truly complete test, we should also verify that all other state variables remain the same
	*/
	Chip8::VirtualMachine::State dummyState;
	
	m_program.bytecode.clear();
	m_program.bytecode.push_back(0x00);
	m_program.bytecode.push_back(0xE0);

	//memcpy(&dummyState.memory[0x200], &m_program.bytecode[0], 2);

	m_vm.load(&m_program);
	dummyState = m_vm.saveState();
	//memset(&dummyState, 0, sizeof dummyState);
	memset(&dummyState.videoBuffer, 0x45, Chip8::VideoBufferSizeMegaChip);

	m_vm.loadState(dummyState);

	m_vm.step(1);

	dummyState = m_vm.saveState();

	for (int i = 0; i < Chip8::VideoBufferSizeMegaChip; ++i)
	{
		if (dummyState.videoBuffer[i] != 0)
			return false;
	}

	return true;
}

bool TestSuiteChip8VM::testReturnFromSubroutine()
{
	return false;
}

bool TestSuiteChip8VM::testJump_NNN()
{
	Chip8::VirtualMachine::State testState;

	m_program.bytecode.clear();
	//0x1NNN
	m_program.bytecode.push_back(0x13);
	m_program.bytecode.push_back(0x33);
	m_vm.load(&m_program);
	m_vm.step(1);
	testState = m_vm.saveState();

#ifdef CHIP8_TESTS_ASSERT
	assert(testState.PC == 0x333);
#endif
	if (testState.PC != 0x333)
		return false;

	return true;
}
								
bool TestSuiteChip8VM::testCall_NNN()
{
	Chip8::VirtualMachine::State testState;
	
	m_program.bytecode.clear();
	//0x2333, subroutine call to memory position 0x333
	m_program.bytecode.push_back(0x23);
	m_program.bytecode.push_back(0x33);

	m_vm.load(&m_program);
	
	m_vm.step(1);

	testState = m_vm.saveState();

#ifdef CHIP8_TESTS_ASSERT
	assert(testState.PC == 0x333);
	assert(testState.stack[0] == 0x200);
	assert(testState.stackPointer = 1);
#endif

	if (testState.PC != 0x333 || testState.stack[0] != 0x200 || testState.stackPointer != 1)
		return false;

	return true;
}
								
bool TestSuiteChip8VM::testSkipNextIf_Vx_Equal_KK()
{
	Chip8::VirtualMachine::State testState;

	//0x3XKK
	m_program.bytecode.clear();
	m_program.bytecode.push_back(0x30);
	m_program.bytecode.push_back(0x00);

	m_vm.load(&m_program);
	m_vm.step(1);
	testState = m_vm.saveState();

#ifdef CHIP8_TESTS_ASSERT
	assert(testState.PC == 0x204);
#endif
	if (testState.PC != 0x204)
		return false;

	testState.PC = 0x200;
	testState.registerV[0] = 1;
	m_vm.loadState(testState);
	m_vm.step(1);
	testState = m_vm.saveState();

#ifdef CHIP8_TESTS_ASSERT
	assert(testState.PC == 0x202);
#endif
	if (testState.PC != 0x202)
		return false;

	return true;
}
					
bool TestSuiteChip8VM::testSkipNextIf_Vx_NotEqual_KK()
{
	Chip8::VirtualMachine::State testState;

	//0x4XKK
	m_program.bytecode.clear();
	m_program.bytecode.push_back(0x40);
	m_program.bytecode.push_back(0x00);

	m_vm.load(&m_program);
	m_vm.step(1);
	testState = m_vm.saveState();

#ifdef CHIP8_TESTS_ASSERT
	assert(testState.PC == 0x202);
#endif
	if (testState.PC != 0x202)
		return false;

	testState.PC = 0x200;
	testState.registerV[0] = 1;
	m_vm.loadState(testState);
	m_vm.step(1);
	testState = m_vm.saveState();

#ifdef CHIP8_TESTS_ASSERT
	assert(testState.PC == 0x204);
#endif
	if (testState.PC != 0x204)
		return false;

	return true;
}
				
bool TestSuiteChip8VM::testSkipNextIf_Vx_Equal_Vy()
{
	Chip8::VirtualMachine::State testState;
	//0x5XY0
	m_program.bytecode.clear();
	m_program.bytecode.push_back(0x50);
	m_program.bytecode.push_back(0x10);

	m_vm.load(&m_program);
	m_vm.step(1);
	testState = m_vm.saveState();

#ifdef CHIP8_TESTS_ASSERT
	assert(testState.PC == 0x204);
#endif
	if (testState.PC != 0x204)
		return false;

	testState.PC = 0x200;
	testState.registerV[1] = 1;
	m_vm.loadState(testState);
	m_vm.step(1);
	testState = m_vm.saveState();

#ifdef CHIP8_TESTS_ASSERT
	assert(testState.PC == 0x202);
#endif
	if (testState.PC != 0x202)
		return false;


	return true;
}
					
bool TestSuiteChip8VM::testVx_Assign_KK()
{
	Chip8::VirtualMachine::State testState;

	for (int x = 0; x < 16; ++x)
	{
		m_program.bytecode.clear();
		m_program.bytecode.push_back(0x60 + x);
		m_program.bytecode.push_back(0x42);
		m_vm.load(&m_program);
		m_vm.step(1);

		testState = m_vm.saveState();

		if (testState.registerV[x] != 0x42)
			return false;
	}

	return true;
}
							
bool TestSuiteChip8VM::testVx_Assign_Vx_Plus_KK()
{
	Chip8::VirtualMachine::State testState;

	m_program.bytecode.clear();
	m_program.bytecode.push_back(0x70);
	m_program.bytecode.push_back(0x01);

	m_vm.load(&m_program);
	testState = m_vm.saveState();
	testState.registerV[0] = 1;

	m_vm.loadState(testState);
	m_vm.step(1);

	testState = m_vm.saveState();

#ifdef CHIP8_TESTS_ASSERT
	assert(testState.registerV[0] == 2);
#endif

	if (testState.registerV[0] != 2)
		return false;
	return true;
}
					

bool TestSuiteChip8VM::testVx_Assign_Vy()
{
	Chip8::VirtualMachine::State testState;

	for (int x = 0; x < 16; ++x)
	{
		int y = x ? 0 : 1;
		m_program.bytecode.clear();
		m_program.bytecode.push_back(0x80 + x);
		m_program.bytecode.push_back(0x00 + (y << 4));
		m_vm.load(&m_program);
		
		//grab the state after loading our test program and modify register Y with a known value
		testState = m_vm.saveState();
		testState.registerV[y] = 0x42;
		m_vm.loadState(testState);

		m_vm.step(1);

		testState = m_vm.saveState();

		if (testState.registerV[x] != 0x42)
			return false;
	}

	return true;
}
							
bool TestSuiteChip8VM::testVx_Assign_Vx_OR_Vy()
{
	Chip8::VirtualMachine::State testState;

	m_program.bytecode.clear();
	m_program.bytecode.push_back(0x80);
	m_program.bytecode.push_back(0x11);
	m_vm.load(&m_program);
	
	testState = m_vm.saveState();
	testState.registerV[0] = 5;
	testState.registerV[1] = 33;
	m_vm.loadState(testState);
	m_vm.step(1);
	testState = m_vm.saveState();

#ifdef CHIP8_TESTS_ASSERT
	assert(testState.registerV[0] == 37);
	assert(testState.registerV[1] == 33);
#endif
	if (testState.registerV[0] != 37 || testState.registerV[1] != 33)
		return false;

	return true;
}

bool TestSuiteChip8VM::testVx_Assign_Vx_AND_Vy()
{
	Chip8::VirtualMachine::State testState;

	m_program.bytecode.clear();
	//V0 AND V1 (0x8012)
	m_program.bytecode.push_back(0x80);
	m_program.bytecode.push_back(0x12);

	/*m_program.bytecode.push_back(0x82);
	m_program.bytecode.push_back(0x32);

	m_program.bytecode.push_back(0x84);
	m_program.bytecode.push_back(0x52);*/
	m_vm.load(&m_program);
	testState = m_vm.saveState();

	//testState.registerV[0] = 1;
	//testState.registerV[1] = 1;

	testState.registerV[0] = 143;
	testState.registerV[1] = 169;
	m_vm.loadState(testState);
	m_vm.step(1);
	testState = m_vm.saveState();

#ifdef CHIP8_TESTS_ASSERT
	assert(testState.registerV[0] == 137);
#endif

	if (testState.registerV[0] != 137)
		return false;

	return true;
}
						
bool TestSuiteChip8VM::testVx_Assign_Vx_XOR_Vy()
{
	return false;
}

bool TestSuiteChip8VM::testVx_Assign_Vx_Plus_Vy()
{
	Chip8::VirtualMachine::State testState;

	//0x8XY4
	m_program.bytecode.clear();
	m_program.bytecode.push_back(0x80);
	m_program.bytecode.push_back(0x14);
	m_vm.load(&m_program);
	testState = m_vm.saveState();

	testState.registerV[0] = 1;
	testState.registerV[1] = 2;
	m_vm.loadState(testState);
	m_vm.step(1);

	testState = m_vm.saveState();

#ifdef CHIP8_TESTS_ASSERT
	assert(testState.registerV[0] == 3);
	assert(testState.registerV[0xF] == 0);
#endif
	if (testState.registerV[0] != 3 || testState.registerV[0xF] != 0)
		return false;

	testState.PC = 0x200;
	testState.registerV[0] = 250;
	testState.registerV[1] = 7;
	m_vm.loadState(testState);
	m_vm.step(1);
	testState = m_vm.saveState();

#ifdef CHIP8_TESTS_ASSERT
	assert(testState.registerV[0] == 1);
	assert(testState.registerV[0xF] == 1);
#endif
	if (testState.registerV[0] != 1 || testState.registerV[0xF] != 1)
		return false;


	//now, let's see what happens if we use VF as a source in the calculation
	m_program.bytecode.clear();
	m_program.bytecode.push_back(0x80);
	m_program.bytecode.push_back(0xF4);
	m_vm.load(&m_program);
	testState = m_vm.saveState();

	testState.registerV[0] = 12;
	testState.registerV[0xF] = 5;
	m_vm.loadState(testState);
	m_vm.step(1);

	testState = m_vm.saveState();

#ifdef CHIP8_TESTS_ASSERT
	assert(testState.registerV[0] == 17);
	assert(testState.registerV[0xF] == 0);
#endif
	if (testState.registerV[0] != 17 || testState.registerV[0xF] != 0)
		return false;
	return true;
}

bool TestSuiteChip8VM::testVx_Assign_Vx_Minus_Vy()
{
	/*
	0x8XY5 - VX = VX - VY, VF = not borrow

	This test is more involved, by my own choosing. It deos the same calculation for each of the 15 registers, except for VF
	TODO: it may be useful to test VF as well, because VF is supposed to be overwritten with the borrow flag
	*/
	Chip8::VirtualMachine::State testState;

	for (int x = 0; x < 16; ++x)
	{
		int y = x ? 0 : 0xF;
		m_program.bytecode.clear();
		//m_program.bytecode.push_back(0x60 + x);
		//m_program.bytecode.push_back(0x42);
		m_program.bytecode.push_back(0x80 + x);
		m_program.bytecode.push_back(0x05 + (y << 4));
		m_vm.load(&m_program);
		
		//grab the state after loading our test program and modify register Y with a known value
		testState = m_vm.saveState();
		testState.registerV[x] = 0x5;
		testState.registerV[y] = 0x1;
		m_vm.loadState(testState);

		m_vm.step(1);

		testState = m_vm.saveState();

		if (x < 15)
		{
			//borrow bit (reg VF) should be 1 if no borrow was performed, so the test should fail if VF == 0
#ifdef CHIP8_TESTS_ASSERT
			assert((testState.registerV[x] == 0x4));
			assert((testState.registerV[0xF] != 0));
#endif
			if ((testState.registerV[x] != 0x4) || (testState.registerV[0xF] == 0))
				return false;
		} else
		{
			//we are trying to modify register F, which is the carry register for this instruction
		}
	}

	for (int x = 0; x < 16; ++x)
	{
		int y = x ? 0 : 0xF;
		m_program.bytecode.clear();
		//m_program.bytecode.push_back(0x60 + x);
		//m_program.bytecode.push_back(0x42);
		m_program.bytecode.push_back(0x80 + x);
		m_program.bytecode.push_back(0x05 + (y << 4));
		m_vm.load(&m_program);
		
		//grab the state after loading our test program and modify register Y with a known value
		testState = m_vm.saveState();
		testState.registerV[x] = 0x1;
		testState.registerV[y] = 0x5;
		m_vm.loadState(testState);

		m_vm.step(1);

		testState = m_vm.saveState();

		if (x < 15)
		{
#ifdef CHIP8_TESTS_ASSERT
			assert(testState.registerV[x] == 0xFC);
			assert(testState.registerV[0xF] != 1);
#endif
			if ((testState.registerV[x] != 0xFC) || (testState.registerV[0xF] == 1))
				return false;
		} else
		{
			//we are trying to modify register F, which is the carry register for this instruction
		}
	}

	return true;
}



bool TestSuiteChip8VM::testVx_Assign_VxVy_ShiftRightBy1()
{
	//TODO
	/*
	8XY6 - VX = VX (VY in the real COSMAC VM) shr 1
	*/


	Chip8::VirtualMachine::State testState;

	m_program.bytecode.clear();
	m_program.bytecode.push_back(0x80);
	m_program.bytecode.push_back(0x16);

	m_vm.load(&m_program);
	testState = m_vm.saveState();

	testState.correctChip8Shift = false;
	testState.registerV[0] = 139;
	m_vm.loadState(testState);
	m_vm.step(1);
	testState = m_vm.saveState();

#ifdef CHIP8_TESTS_ASSERT
	assert(testState.registerV[0] == 69);
	assert(testState.registerV[0xF] == 1);
#endif
	if (testState.registerV[0] != 69 || testState.registerV[0xF] != 1)
		return false;

	m_program.bytecode.clear();
	m_program.bytecode.push_back(0x80);
	m_program.bytecode.push_back(0xF6);

	m_vm.load(&m_program);
	testState = m_vm.saveState();

	testState.correctChip8Shift = true;
	testState.registerV[0xF] = 139;
	m_vm.loadState(testState);
	m_vm.step(1);
	testState = m_vm.saveState();

#ifdef CHIP8_TESTS_ASSERT
	assert(testState.registerV[0] == 69);
	assert(testState.registerV[0xF] == 1);
#endif
	if (testState.registerV[0] != 69 || testState.registerV[0xF] != 1)
		return false;

	return true;
}
			
bool TestSuiteChip8VM::testVx_Assign_Vy_Minus_Vx()
{
	/*
	0x8XY7 - VX = VY - VX, VF = not borrow
	*/
	Chip8::VirtualMachine::State testState;

	m_program.bytecode.clear();
	m_program.bytecode.push_back(0x80);
	m_program.bytecode.push_back(0xF7);
	m_vm.load(&m_program);

	testState = m_vm.saveState();
	//first we test 5 - 1 = 4, which does not need to borrow, so VF should equal 1
	testState.registerV[0] = 1;
	testState.registerV[0xF] = 1;
	m_vm.loadState(testState);
	m_vm.step(1);

	testState = m_vm.saveState();

#ifdef CHIP8_TESTS_ASSERT
	assert(testState.registerV[0] == 0);
	assert(testState.registerV[0xF] == 1);
#endif
	if (testState.registerV[0] != 0 || testState.registerV[0xF] != 1)
		return false;

	//test 1 - 5 = 252, this needs borrow, so VF should equal 0
	testState.PC = 0x200;
	testState.registerV[0] = 5;
	testState.registerV[0xF] = 1;
	m_vm.loadState(testState);
	m_vm.step(1);
	testState = m_vm.saveState();

#ifdef CHIP8_TESTS_ASSERT
	assert(testState.registerV[0] == 0xFC);
	assert(testState.registerV[0xF] == 0);
#endif
	if (testState.registerV[0] != 0xFC || testState.registerV[0xF] != 0)
		return false;

	return true;
}


bool TestSuiteChip8VM::testVx_Assign_VxVy_ShiftLeftBy1()
{
	Chip8::VirtualMachine::State testState;

	//0x8XYE
	//for the incorrect shift (vx = vx shr 1), we should just not allow VF being used, because it makes no sense
	//this may be needed though, if something dependson this weird behaviour
	m_program.bytecode.clear();
	m_program.bytecode.push_back(0x80);
	m_program.bytecode.push_back(0x1E);

	m_vm.load(&m_program);
	testState = m_vm.saveState();

	testState.correctChip8Shift = false;
	testState.registerV[0] = 139;
	m_vm.loadState(testState);
	m_vm.step(1);
	testState = m_vm.saveState();

#ifdef CHIP8_TESTS_ASSERT
	assert(testState.registerV[0] == 22);
	assert(testState.registerV[0xF] == 1);
#endif
	if (testState.registerV[0] != 22 || testState.registerV[0xF] != 1)
		return false;

	m_program.bytecode.clear();
	m_program.bytecode.push_back(0x80);
	m_program.bytecode.push_back(0xFE);

	m_vm.load(&m_program);
	testState = m_vm.saveState();

	testState.correctChip8Shift = true;
	testState.registerV[0xF] = 139;
	m_vm.loadState(testState);
	m_vm.step(1);
	testState = m_vm.saveState();

#ifdef CHIP8_TESTS_ASSERT
	assert(testState.registerV[0] == 22);
	assert(testState.registerV[0xF] == 1);
#endif
	if (testState.registerV[0] != 22 || testState.registerV[0xF] != 1)
		return false;

	return true;
}
				

bool TestSuiteChip8VM::testSkipNextIf_Vx_NotEqual_Vy()
{
	return false;
}
				
bool TestSuiteChip8VM::testI_Assign_NNN()
{
	Chip8::VirtualMachine::State testState;

	//0xANNN, NNN is an address into memory, register I should be set to this value
	m_program.bytecode.clear();
	m_program.bytecode.push_back(0xA3);
	m_program.bytecode.push_back(0x22);
	m_vm.load(&m_program);

	m_vm.step(1);

	testState = m_vm.saveState();

#ifdef CHIP8_TESTS_ASSERT
	assert(testState.registerI == 0x322);
#endif
	if (testState.registerI != 0x322)
		return false;

	return true;
}
							
bool TestSuiteChip8VM::testJumpTo_NNN_Plus_V0()
{
	return false;
}
						
bool TestSuiteChip8VM::testVx_Assign_Random_AND_KK()
{
	Chip8::VirtualMachine::State testState;

	m_program.bytecode.clear();
	m_program.bytecode.push_back(0xC0);
	m_program.bytecode.push_back(0x30);

	m_vm.load(&m_program);

	int samples[0x31];
	memset(samples, 0, sizeof(int)*0x31);

	for (int i = 0; i < 100; ++i)
	{
		m_vm.step(1);

		testState = m_vm.saveState();
		
#ifdef CHIP8_TESTS_ASSERT
		assert(testState.registerV[0] <= 0x30);
#endif
		if (testState.registerV[0] > 0x30)
			return false;

		//some crude testing to see if the random generation produces too many of the same numbers
		samples[ testState.registerV[0] ]++;
#ifdef CHIP8_TESTS_ASSERT
		assert(samples[ testState.registerV[0] ] < 50);
#endif
		if (samples[ testState.registerV[0] ] >= 50)
			return false;
			
		testState.PC = 0x200;
		testState.registerV[0] = 0;
		m_vm.loadState(testState);
	}

	return true;
}
	
static const char Sprite85Data[] = { 
	1, 1, 1, 1, 0, 0, 0, 0,
	1, 0, 0, 1, 0, 0, 0, 0,
	1, 0, 0, 1, 0, 0, 0, 0,
	1, 0, 0, 1, 0, 0, 0, 0,
	1, 1, 1, 1, 0, 0, 0, 0
};

static const char Sprite16Data[] = { 
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
	1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
	1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
	1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1,
	1, 0, 0, 0, 1, 0 ,0 ,0 ,0, 0, 0, 1, 0, 0, 0, 1,
	1, 0, 0, 0, 1, 0 ,0 ,0 ,0, 0, 0, 1, 0, 0, 0, 1,
	1, 0, 0, 0, 1, 0 ,0 ,0 ,0, 0, 0, 1, 0, 0, 0, 1,
	1, 0, 0, 0, 1, 0 ,0 ,0 ,0, 0, 0, 1, 0, 0, 0, 1,
	1, 0, 0, 0, 1, 0 ,0 ,0 ,0, 0, 0, 1, 0, 0, 0, 1,
	1, 0, 0, 0, 1, 0 ,0 ,0 ,0, 0, 0, 1, 0, 0, 0, 1,
	1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1,
	1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
	1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
	1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
};
bool TestSuiteChip8VM::testDrawSprite()
{
	/*

	DXYN
	XY, indicate registerV[x] and registerV[y] that contain the x,y coordinate for the sprite
	N indicate how many lines to draw, which in chip8 mode equals how many bytes to draw
	in schip, it equals how many bytes to draw * 2

	Register I has to be set to the memory location of the sprite to draw.
	*/


	Chip8::VirtualMachine::State testState;

	m_program.bytecode.clear();
	m_program.bytecode.push_back(0xD0);
	m_program.bytecode.push_back(0x15);
	m_vm.load(&m_program);

	/*
	test 8xN sprite
	*/
	testState = m_vm.saveState();
	testState.registerV[0] = 0;
	testState.registerV[1] = 0;
	//drawing data, the number 0 copied from the internal VM data into here
	testState.memory[0x300] = 0xF0;
	testState.memory[0x301] = 0x90;
	testState.memory[0x302] = 0x90;
	testState.memory[0x303] = 0x90;
	testState.memory[0x304] = 0xF0;
	testState.registerI = 0x300;

	m_vm.loadState(testState);
	m_vm.step(1);
	testState = m_vm.saveState();

	for (int y = 0; y < 5; ++y)
	{
		for (int x = 0; x < 8; ++x)
		{
			int xpos = x * 2;
			int ypos = y * 2;
#ifdef CHIP8_TESTS_ASSERT
			assert(testState.videoBuffer[xpos + ypos*128] == Sprite85Data[x + y*8]);
#endif
			
			if (testState.videoBuffer[xpos + ypos*128] != Sprite85Data[x + y*8] &&
				testState.videoBuffer[xpos + 1 + (ypos+1)*128] != Sprite85Data[x + y*8] &&
				testState.videoBuffer[xpos + 1 + ypos*128] != Sprite85Data[x + y*8] &&
				testState.videoBuffer[xpos + (ypos+1)*128] != Sprite85Data[x + y*8])
			{
				return false;
			}
		}
	}


	
	/*
	test 8x16 sprite
	*/
	m_program.bytecode.clear();
	m_program.bytecode.push_back(0xD0);
	m_program.bytecode.push_back(0x10);
	m_vm.load(&m_program);

	testState = m_vm.saveState();
	testState.registerV[0] = 0;
	testState.registerV[1] = 0;
	//drawing data, the number 0 copied from the internal VM data into here
	testState.memory[0x300] = 0xFF;	testState.memory[0x301] = 0xFF;	testState.memory[0x302] = 0xFF;	testState.memory[0x303] = 0xFF;
	testState.memory[0x304] = 0xFF; testState.memory[0x305] = 0xFF; testState.memory[0x306] = 0xFF; testState.memory[0x307] = 0xFF;
	testState.memory[0x308] = 0xFF; testState.memory[0x309] = 0xFF; testState.memory[0x30A] = 0xFF; testState.memory[0x30B] = 0xFF;
	testState.memory[0x30C] = 0xFF; testState.memory[0x30D] = 0xFF; testState.memory[0x30E] = 0xFF; testState.memory[0x30F] = 0xFF;

	testState.registerI = 0x300;

	m_vm.loadState(testState);
	m_vm.step(1);
	testState = m_vm.saveState();

	for (int y = 0; y < 16; ++y)
	{
		for (int x = 0; x < 8; ++x)
		{
			int xpos = x * 2;
			int ypos = y * 2;
#ifdef CHIP8_TESTS_ASSERT
			assert(testState.videoBuffer[xpos + ypos*128] == 1);
#endif
			if (testState.videoBuffer[xpos + ypos*128] != 1)
			{
				return false;
			}
		}
	}

	//test 16x16 sprite

	m_program.bytecode.clear();
	m_program.bytecode.push_back(0xD0);
	m_program.bytecode.push_back(0x10);
	m_vm.load(&m_program);

	testState = m_vm.saveState();
	testState.registerV[0] = 0;
	testState.registerV[1] = 0;
	//drawing data, the number 0 copied from the internal VM data into here
	testState.memory[0x300] = 0xFF;	testState.memory[0x301] = 0xFF;	
	testState.memory[0x302] = 0x80;	testState.memory[0x303] = 0x01;
	testState.memory[0x304] = 0x80; testState.memory[0x305] = 0x01; 
	testState.memory[0x306] = 0x80; testState.memory[0x307] = 0x01;
	testState.memory[0x308] = 0x8F; testState.memory[0x309] = 0xF1; 
	testState.memory[0x30A] = 0x88; testState.memory[0x30B] = 0x11;
	testState.memory[0x30C] = 0x88; testState.memory[0x30D] = 0x11; 
	testState.memory[0x30E] = 0x88; testState.memory[0x30F] = 0x11;

	testState.memory[0x310] = 0x88;	testState.memory[0x311] = 0x11;	
	testState.memory[0x312] = 0x88;	testState.memory[0x313] = 0x11;
	testState.memory[0x314] = 0x88; testState.memory[0x315] = 0x11; 
	testState.memory[0x316] = 0x8F; testState.memory[0x317] = 0xF1;
	testState.memory[0x318] = 0x80; testState.memory[0x319] = 0x01; 
	testState.memory[0x31A] = 0x80; testState.memory[0x31B] = 0x01;
	testState.memory[0x31C] = 0x80; testState.memory[0x31D] = 0x01; 
	testState.memory[0x31E] = 0xFF; testState.memory[0x31F] = 0xFF;

	testState.registerI = 0x300;
	testState.superChipMode = true;

	m_vm.loadState(testState);
	m_vm.step(1);
	testState = m_vm.saveState();

	for (int y = 0; y < 16; ++y)
	{
		for (int x = 0; x < 16; ++x)
		{
#ifdef CHIP8_TESTS_ASSERT
			assert(testState.videoBuffer[x + y*128] == Sprite16Data[x + y*16]);
#endif
			if (testState.videoBuffer[x + y*128] != Sprite16Data[x + y*16])
			{
				return false;
			}
		}
	}


	//TODO: add megachip drawing test
	//...or possibly do that in a separate test method, that would make more sense to avoid having too much code in here
	return true;
}
							
bool TestSuiteChip8VM::testSkipNextIf_Key_Vx_Pressed()
{
	return false;
}
			
bool TestSuiteChip8VM::testSkipNextIf_Key_Vx_NotPressed()
{
	return false;
}

bool TestSuiteChip8VM::testI_Assign_NN()
{
	return false;
}

bool TestSuiteChip8VM::testVx_Assign_DelayTimer()
{
	return false;
}

bool TestSuiteChip8VM::testWaitForKeypress_StoreKeyIn_Vx()
{
	return false;
}

bool TestSuiteChip8VM::testDelayTimer_Assign_Vx()
{
	return false;
}

bool TestSuiteChip8VM::testSoundTimer_Assign_Vx()
{
	return false;
}

bool TestSuiteChip8VM::testI_Assign_I_Plus_Vx()
{
	Chip8::VirtualMachine::State testState;

	//0xF01E, I += regV[0]
	m_program.bytecode.clear();
	m_program.bytecode.push_back(0xF0);
	m_program.bytecode.push_back(0x1E);
	m_vm.load(&m_program);

	testState = m_vm.saveState();

	testState.registerI = 5;
	testState.registerV[0] = 4;
	m_vm.loadState(testState);
	m_vm.step(1);
	testState = m_vm.saveState();

#ifdef CHIP8_TESTS_ASSERT
	assert(testState.registerI == 9);
#endif
	if (testState.registerI != 9)
		return false;

	return true;
}

bool TestSuiteChip8VM::testI_Assign_FontSpriteForCharacterIn_Vx()
{
	return false;
}

bool TestSuiteChip8VM::testStoreBCD_Of_Vx_InMemory()
{
	return false;
}

bool TestSuiteChip8VM::testStoreV0_To_Vx_InMemory()
{
	return false;
}

bool TestSuiteChip8VM::testReadV0_To_Vx_FromMemory()
{
	Chip8::VirtualMachine::State testState;

	/*
	FX65 - Read V0 to VX from the memory, register I points out the starting address in memory
	There is some confusion whether register I should be incremented by X after
	Currently we assume that it should be, but some documentation and people state otherwise
	*/
	m_program.bytecode.clear();
	m_program.bytecode.push_back(0xFF);
	m_program.bytecode.push_back(0x65);

	m_vm.load(&m_program);

	testState = m_vm.saveState();
	//set data for the test case
	testState.registerI = 0x300;
	memset(&testState.memory[0x300], 0x42, 16);
	m_vm.loadState(testState);

	m_vm.step(1);

	testState = m_vm.saveState();
//The Chip8VM should increase I, and thus we should verify that it does here, but apparently some games actually depend on I not increasing with this
//instruction, so because of that, we can't do the correct thing
//#ifdef CHIP8_TESTS_ASSERT
//	assert(testState.registerI == 0x30F);
//#endif
//	if (testState.registerI != 0x30F)
//		return false;

	for (int i = 0; i < 16; ++i)
	{
		
#ifdef CHIP8_TESTS_ASSERT
		assert(testState.registerV[i] == 0x42);
#endif
		if (testState.registerV[i] != 0x42)
			return false;
	}
		
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//SuperChip8 tests

bool TestSuiteChip8VM::testScrollDown_N_Lines()
{
	Chip8::VirtualMachine::State testState;

	m_program.bytecode.clear();
	m_program.bytecode.push_back(0x00);
	m_program.bytecode.push_back(0xC2);
	m_vm.load(&m_program);
	testState = m_vm.saveState();

	for (int y = 0; y < 16; ++y)
	{
		memcpy(&testState.videoBuffer[y * 128], &Sprite16Data[y * 16], 16);
	}
	
	m_vm.loadState(testState);
	m_vm.step(1);
	testState = m_vm.saveState();

	for (int i = 0; i < 64; ++i)
	{
#ifdef CHIP8_TESTS_ASSERT
	assert(testState.videoBuffer[i] == 0);
#endif
		if (testState.videoBuffer[i] != 0)
			return false;
	}

	for (int y = 0; y < 16; ++y)
	{
		for (int x = 0; x < 16; ++x)
		{
#ifdef CHIP8_TESTS_ASSERT
	assert(testState.videoBuffer[x + (y + 2)*128] == Sprite16Data[x + y*16]);
#endif
			//testState.videoBuffer[ x + y*16] = Sprite16Data[x + y*16];
			if (testState.videoBuffer[ x + (y + 2)*128] != Sprite16Data[x + y*16])
				return false;
		}
	}

	/*
	now let's try it in superchip mode as well
	*/
	m_vm.load(&m_program);
	testState = m_vm.saveState();
	testState.superChipMode = true;

	for (int y = 0; y < 16; ++y)
	{
		memcpy(&testState.videoBuffer[y * 128], &Sprite16Data[y * 16], 16);
	}

	m_vm.loadState(testState);
	m_vm.step(1);
	testState = m_vm.saveState();

	for (int i = 0; i < 128; ++i)
	{
#ifdef CHIP8_TESTS_ASSERT
		assert(testState.videoBuffer[i] == 0);
#endif
		if (testState.videoBuffer[i] != 0)
			return false;
	}

	for (int y = 0; y < 16; ++y)
	{
		for (int x = 0; x < 16; ++x)
		{
#ifdef CHIP8_TESTS_ASSERT
			assert(testState.videoBuffer[x + (y + 2)*128] == Sprite16Data[x + y*16]);
#endif
			//testState.videoBuffer[ x + y*16] = Sprite16Data[x + y*16];
			if (testState.videoBuffer[ x + (y + 2)*128] != Sprite16Data[x + y*16])
				return false;
		}
	}
	
	//TODO: add test for half-pixel shift
	return true;
}

bool TestSuiteChip8VM::testScroll_Right()
{
	/*
	0x00FB, scroll 4 pixels to the right, should be zeroed at the start and pixels moved out of bounds should disappear
	*/
	Chip8::VirtualMachine::State testState;

	m_program.bytecode.clear();
	m_program.bytecode.push_back(0x00);
	m_program.bytecode.push_back(0xFB);
	m_vm.load(&m_program);
	testState = m_vm.saveState();

	for (int y = 0; y < 16; ++y)
	{
		memcpy(&testState.videoBuffer[y * 128], &Sprite16Data[y * 16], 16);
	}

	m_vm.loadState(testState);
	m_vm.step(1);
	testState = m_vm.saveState();

	for (int y = 0; y < 16; ++y)
	{
		for (int x = 0; x < 4; ++x)
		{
#ifdef CHIP8_TESTS_ASSERT
			assert(testState.videoBuffer[x + y * 128] == 0);
#endif
			if (testState.videoBuffer[x + y * 128] != 0)
				return false;
		}
	}

	for (int y = 0; y < 16; ++y)
	{
		for (int x = 0; x < 16; ++x)
		{
#ifdef CHIP8_TESTS_ASSERT
			assert(testState.videoBuffer[4 + x + y*128] == Sprite16Data[x + y*16]);
#endif
			if (testState.videoBuffer[4 + x + y*128] != Sprite16Data[x + y*16])
				return false;
		}
	}


	return true;
}
bool TestSuiteChip8VM::testScroll_Left()
{
	/*
	0x00FB, scroll 4 pixels to the right, should be zeroed at the start and pixels moved out of bounds should disappear
	*/
	Chip8::VirtualMachine::State testState;

	m_program.bytecode.clear();
	m_program.bytecode.push_back(0x00);
	m_program.bytecode.push_back(0xFC);
	m_vm.load(&m_program);
	testState = m_vm.saveState();

	for (int y = 0; y < 16; ++y)
	{
		memcpy(&testState.videoBuffer[4 + y * 128], &Sprite16Data[y * 16], 16);
	}

	m_vm.loadState(testState);
	m_vm.step(1);
	testState = m_vm.saveState();

	for (int y = 0; y < 16; ++y)
	{
		for (int x = 16; x < 20; ++x)
		{
#ifdef CHIP8_TESTS_ASSERT
			assert(testState.videoBuffer[x + y * 128] == 0);
#endif
			if (testState.videoBuffer[x + y * 128] != 0)
				return false;
		}
	}

	for (int y = 0; y < 16; ++y)
	{
		for (int x = 0; x < 16; ++x)
		{
#ifdef CHIP8_TESTS_ASSERT
			assert(testState.videoBuffer[x + y*128] == Sprite16Data[x + y*16]);
#endif
			if (testState.videoBuffer[x + y*128] != Sprite16Data[x + y*16])
				return false;
		}
	}


	return true;
}
bool TestSuiteChip8VM::testExit()
{
	Chip8::VirtualMachine::State testState;

	m_program.bytecode.clear();
	m_program.bytecode.push_back(0x00);
	m_program.bytecode.push_back(0xFD);
	m_vm.load(&m_program);
	m_vm.step(1);

#ifdef CHIP8_TESTS_ASSERT
	assert(m_vm.getExecuteStatus() == Chip8::VirtualMachine::Status::Stopped);
#endif
	if (m_vm.getExecuteStatus() != Chip8::VirtualMachine::Status::Stopped)
		return false;

	//now lets verify that the VM does not try to keep going after it's stopped
	m_vm.step(1);
	testState = m_vm.saveState();
#ifdef CHIP8_TESTS_ASSERT
	assert(testState.PC == 0x200);
#endif
	if (testState.PC != 0x200)
		return false;

	return true;
}
bool TestSuiteChip8VM::testDisableSuperChipMode()
{

	return false;
}
bool TestSuiteChip8VM::testEnableSuperChipMode()
{

	return false;
}

bool TestSuiteChip8VM::testStoreV0_To_Vx_InHP48()
{
	/*
	FX75 - store V0 to VX in the special HP48 registers (from the HP48 graph calculator which was the main target of SuperChip I think?).
	They are generally referred to as the HP48 user flags, does that name make sense?
	X only goes up to 7
	*/
	Chip8::VirtualMachine::State testState;

	m_program.bytecode.clear();
	m_program.bytecode.push_back(0xF7);
	m_program.bytecode.push_back(0x75);
	m_vm.load(&m_program);
	
	testState = m_vm.saveState();
	for (int i = 0; i < 8; ++i)
		testState.registerV[i] = i;

	m_vm.loadState(testState);
	m_vm.step(1);
	testState = m_vm.saveState();

	for (int i = 0; i < 8; ++i)
	{
#ifdef CHIP8_TESTS_ASSERT
		assert(testState.HP48UserFlag[i] == i);
#endif
		if (testState.HP48UserFlag[i] != i)
			return false;
	}

	return true;
}

bool TestSuiteChip8VM::testReadV0_To_Vx_FromHP48()
{
	/*
	FX85 - load V0 to VX from the special HP48 registers (from the HP48 graph calculator which was the main target of SuperChip I think?).
	They are generally referred to as the HP48 user flags, does that name make sense?
	X only goes up to 7
	*/
	Chip8::VirtualMachine::State testState;

	m_program.bytecode.clear();
	m_program.bytecode.push_back(0xF7);
	m_program.bytecode.push_back(0x85);
	m_vm.load(&m_program);

	testState = m_vm.saveState();
	for (int i = 0; i < 8; ++i)
		testState.HP48UserFlag[i] = i;

	m_vm.loadState(testState);
	m_vm.step(1);
	testState = m_vm.saveState();

	for (int i = 0; i < 8; ++i)
	{
#ifdef CHIP8_TESTS_ASSERT
		assert(testState.registerV[i] == i);
#endif
		if (testState.registerV[i] != i)
			return false;
	}

	return true;
	return false;
}