#include "Chip8VM.hpp"

/*
TODO: add checks on all math ops to verify that VF carry isn't being set before the actual math op
*/

class TestSuiteChip8VM
{
	public:

		bool runAll();

		void testALU();


		//chip8
		bool testClearDisplay();							//00E0
		bool testReturnFromSubroutine();					//00EE
		bool testJump_NNN();								//1NNN
		bool testCall_NNN();								//2NNN
		bool testSkipNextIf_Vx_Equal_KK();					//3XKK
		bool testSkipNextIf_Vx_NotEqual_KK();				//4XKK
		bool testSkipNextIf_Vx_Equal_Vy();					//5XY0
		bool testVx_Assign_KK();							//6XKK
		bool testVx_Assign_Vx_Plus_KK();					//7XKK

		bool testVx_Assign_Vy();							//8XY0
		bool testVx_Assign_Vx_OR_Vy();						//8XY1
		bool testVx_Assign_Vx_AND_Vy();						//8XY2
		bool testVx_Assign_Vx_XOR_Vy();						//8XY3
		bool testVx_Assign_Vx_Plus_Vy();					//8XY4
		bool testVx_Assign_Vx_Minus_Vy();					//8XY5
		/*
		This is one of the troublesome shift instructions. It is correct to the original Chip8 to shift Vy, however, it seems almost all games need it to do the
		shift on Vx instead. This was incorrectly reported in all documentation until just a few years back, so all emulators did it wrong.
		
		To be complete, the tests for the shift instructions will test both methods (shifting VX and VY respectively).
		*/
		bool testVx_Assign_VxVy_ShiftRightBy1();			//8XY6
		bool testVx_Assign_Vy_Minus_Vx();					//8XY7
		//The second shift instruction, same issue with wrong register being used.
		bool testVx_Assign_VxVy_ShiftLeftBy1();				//8XYE

		bool testSkipNextIf_Vx_NotEqual_Vy();				//9XY0
		bool testI_Assign_NNN();							//ANNN
		bool testJumpTo_NNN_Plus_V0();						//BNNN			# not implemented in VM yet, used by Sokoban
		bool testVx_Assign_Random_AND_KK();					//CXKK
		bool testDrawSprite();								//DXYN			# done, but still not entirely sure we got this right --- revisit to verify, changing chip8 res
		bool testSkipNextIf_Key_Vx_Pressed();				//EX9E			# ant
		bool testSkipNextIf_Key_Vx_NotPressed();			//EXA1			# ant - this is the one in the sample data, it triggered the sequence correctly, but the test has not been built yet, so it is relevant

		bool testI_Assign_NN();								//F0NN			# not implemented in VM yet
		bool testVx_Assign_DelayTimer();					//FX07			#
		bool testWaitForKeypress_StoreKeyIn_Vx();			//FX0A			#
		bool testDelayTimer_Assign_Vx();					//FX15			#
		bool testSoundTimer_Assign_Vx();					//FX18			#
		bool testI_Assign_I_Plus_Vx();						//FX1E
		bool testI_Assign_FontSpriteForCharacterIn_Vx();	//FX29/FX30		#
		bool testStoreBCD_Of_Vx_InMemory();					//FX33			#
		/*
		See comments in definition for more info on register I handling with the store/load ops
		*/
		bool testStoreV0_To_Vx_InMemory();					//FX55			#
		bool testReadV0_To_Vx_FromMemory();					//FX65			# ant

		//superchip
		bool testScrollDown_N_Lines();						//00CN			# revisit to verify, changing chip8 res
		bool testScroll_Right();							//00FB
		bool testScroll_Left();								//00FC
		bool testExit();									//00FD			#
		bool testDisableSuperChipMode();					//00FE			#
		bool testEnableSuperChipMode();						//00FF			#
		bool testStoreV0_To_Vx_InHP48();					//FX75
		bool testReadV0_To_Vx_FromHP48();					//FX85

		//megachip
		bool testDisableMegaChipMode();						//0010
		bool testEnableMegaChipMode();						//0011
		bool testI_Assign_NNNNNN();							//01NN, needs investigation
		bool testLoadColorPalette_I();						//02nn
		bool testSetSpriteWidth_NN();
		bool testSetSpriteHeight_NN();
		bool testSetScreenAlpha_NN();
		bool testPlaySound_I();
		bool testStopSound();
		bool testSetSpriteBlendMode();
		bool testScrollUp_N_Lines();

	private:
		Chip8::Program			m_program;
		Chip8::VirtualMachine	m_vm;
};