/**

NOTE:
There is a crash bug that will invariably occur after 60-90 secs of running (1m 22sec seem to be the time measured when it crashes),
if the open file dialog has been used.
This is a known bug, http://mfctips.com/2012/10/29/cfiledialogdomodal-causes-access-violation/#disqus_thread
although normally related to MFC, but it seems we get the same issues, only after a different time (MFC crash happens after exactly 60 sec)

When we do a release build, this is supposed to disappear completely.
Update (few hours later): Even release builds will fall to this strange crash bug after a similar amount of time...
So maybe not related to the MFC bug at all. Which would make sense since we already had to change criterias to fit in there, and we don't use MFC.

*/
#define WIN32_LEAN_AND_MEAN
#include <Windows.h>
#include <Wbemidl.h>

#define CHIP8_TESTS_ASSERT

#include "TestChip8VM.hpp"
#include "MainWindow.hpp"
#include "Chip8VM.hpp"
#include "Utility.h"

#include <fstream>
#include <iomanip>

/**
Chip8 emulator Logo [Garstyciuks].ch8

copied in here to provide a built-in start screen
*/
const char Chip8EmulatorLogo[] = {
0x0, 0xE0, 
0x60, 0x0, 0x61, 0x0, 0x62, 0x8, 0xa2, 0x20, 0x40, 0x40, 0x22, 0x1a, 0x41, 0x20, 0x12, 
0x10, 0xd0, 0x18, 0xf2, 0x1e, 0x70, 0x8, 0x12, 0xa, 0x60, 0x0, 0x71, 0x8, 0x0, 0xee, 
0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 
0x0, 0x0, 0x7f, 0x40, 0x5f, 0x50, 0x57, 0x54, 0x54, 0x0, 0xfc, 0x4, 0xf4, 0x14, 0xd4, 
0x54, 0x54, 0x0, 0x3f, 0x20, 0x2f, 0x28, 0x2b, 0x2a, 0x2a, 0x0, 0xfe, 0x2, 0xfa, 0xa, 
0xea, 0x2a, 0x2a, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 
0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 
0x0, 0x0, 0x0, 0x0, 0x0, 0x54, 0x54, 0x54, 0x54, 0x54, 0x54, 0x74, 0x0, 0x54, 0x54, 
0x54, 0x54, 0x74, 0x0, 0x0, 0x0, 0x2a, 0x2a, 0x2a, 0x2a, 0x2a, 0x2a, 0x3b, 0x0, 0x2a, 
0x2a, 0x2a, 0x2a, 0x2a, 0x2a, 0xee, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 
0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 
0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x74, 0x54, 0x54, 0x54, 0x54, 0x54, 
0x54, 0x54, 0x0, 0x0, 0x74, 0x54, 0x54, 0x54, 0x54, 0x54, 0x3b, 0x2a, 0x2a, 0x2a, 0x2a, 
0x2a, 0x2a, 0x2a, 0xee, 0x2a, 0x2a, 0x2a, 0x2a, 0x2a, 0x2a, 0x2a, 0x0, 0x0, 0x0, 0x0, 
0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 
0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x54, 0x54, 
0x57, 0x50, 0x5f, 0x40, 0x7f, 0x0, 0x54, 0x54, 0xd4, 0x14, 0xf4, 0x4, 0xfc, 0x0, 0x2a, 
0x2a, 0x2b, 0x28, 0x2f, 0x20, 0x3f, 0x0, 0x2a, 0x2a, 0xea, 0xa, 0xfa, 0x2, 0xfe, 0x0, 
0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 
0x0 
};



bool keyStates[sf::Keyboard::KeyCount];

void loadTestFile();
bool isKeyPressed(sf::Keyboard::Key key);
void runTests();

//LONG WINAPI MyAppCrashHandler(EXCEPTION_POINTERS * /*ExceptionInfo*/)
//{
//    //std::cout << "Gotcha!" << std::endl;
//	OutputDebugString("\nCaught shit");
//    return EXCEPTION_CONTINUE_EXECUTION;
//}


int WINAPI wWinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, PWSTR pCmdLine, int nCmdShow)
{
	//init COM
	HRESULT hres;

	//hres =  CoInitializeEx(0, COINIT_APARTMENTTHREADED); 
	//SetUnhandledExceptionFilter(MyAppCrashHandler);
	/*
	Add some debugging features if we're running 64bit Windows 7, since there are some problems causing crashes to be swallowed
	*/
	OSVERSIONINFO osvi;
	SYSTEM_INFO si;
	GetVersionEx(&osvi);
	GetNativeSystemInfo(&si);

	//6.1 indicates windows 7 apparently, sadly I forgot to document where I found this information when this was written
	if (osvi.dwMajorVersion == 6 && osvi.dwMinorVersion == 1 && si.wProcessorArchitecture == PROCESSOR_ARCHITECTURE_AMD64)
	{
		/*
			Thanks to Bruce Dawson for writing this article on the subject:
			http://www.altdevblogaday.com/2012/07/06/when-even-crashing-doesnt-work/
		*/
		typedef BOOL (WINAPI *tGetPolicy)(LPDWORD lpFlags); 
		typedef BOOL (WINAPI *tSetPolicy)(DWORD dwFlags); 
		const DWORD EXCEPTION_SWALLOWING = 0x1;

		HMODULE kernel32 = LoadLibraryA("kernel32.dll");
		//TODO: error checking on "kernel32"
		//it could fail now and we would still proceed as if nothing's wrong
		tGetPolicy pGetPolicy = (tGetPolicy)GetProcAddress(kernel32, "GetProcessUserModeExceptionPolicy"); 
		tSetPolicy pSetPolicy = (tSetPolicy)GetProcAddress(kernel32, "SetProcessUserModeExceptionPolicy"); 
		if (pGetPolicy && pSetPolicy) 
		{ 
			DWORD dwFlags;
			if (pGetPolicy(&dwFlags)) 
			{ 
				// Turn off the filter 
				pSetPolicy(dwFlags & ~EXCEPTION_SWALLOWING); 
			} 
		}
	}

	memset(keyStates, 0, sizeof keyStates);
	//creating a window that is "locally topmost" to my application, needs to be figured out
	/*sf::WindowHandle handle = debugWindow.getSystemHandle();
	RECT rect;
	GetWindowRect(handle, &rect);

	SetWindowPos(handle, HWND_NOTOPMOST, rect.left, rect.top, 200, 200, SWP_SHOWWINDOW);*/

	/*int x = 143, y = 169;
	x = x & y;
	char testm[200];
	sprintf(testm, "\nX & Y: %d", x);
	OutputDebugString(testm);*/

	runTests();

	UI::Default::Initialize(getExecutablePath() + "/resources/luximb.ttf");
	MainWindow wnd;

	MainWindowAppInterface mwai(&wnd);
	Chip8::VirtualMachine vm(&mwai);

	
	Debugger chip8debugger;
	wnd.insertVMPointers(&vm, &chip8debugger);
	vm.attachDebugger(&chip8debugger);

	Chip8::Program program;

	//loadTestFile();
	//program.bytecode.reserve(289);
	program.bytecode = std::vector<uint8_t>(Chip8EmulatorLogo, Chip8EmulatorLogo + 288);
	vm.load(&program);

	sf::Clock clock;
	uint64_t frameTime;
	uint64_t accumulator = 0;

	bool exitProgram = false;
	bool runChip8VM = true;



	while (!exitProgram)
	{
		if (isKeyPressed(sf::Keyboard::Pause))
		{
			runChip8VM = !runChip8VM;
		}
		

		if (runChip8VM)
		{
			//handle the standard timing stuff and grab delta
			frameTime = clock.restart().asMicroseconds();
			//16667 us is 1/60th of a second
			if (frameTime > DesiredFrameTime)
				frameTime = DesiredFrameTime;
		
			accumulator += frameTime;

			while (accumulator >= DesiredFrameTime)
			{
				vm.update(DesiredFrameTime);
				accumulator -= DesiredFrameTime;
			}
		}

		/*wnd.clearBuffer();
		wnd.drawChip8Buffer(vm.getHardwareOutput().videoBuffer, vm.getHardwareOutput().width, vm.getHardwareOutput().height);*/
		exitProgram = wnd.update();
	}

	//cleanup COM
	//CoUninitialize();
}

bool isKeyPressed(sf::Keyboard::Key key)
{
	if (sf::Keyboard::isKeyPressed(key))
	{
		if (keyStates[key] == true)
		{
			return false;
		} else
		{
			keyStates[key] = true;
			return true;
		}
	} else
	{
		keyStates[key] = false;
	}

	return false;
}

void runTests()
{
	TestSuiteChip8VM testc8;
	

	{
		OutputDebugString("\n"); OutputDebugString("testClearDisplay(): ");
		if (testc8.testClearDisplay())
			OutputDebugString("passed");
		else
			OutputDebugString("failed");
	}

	{
		OutputDebugString("\n"); OutputDebugString("testJump_NNN(): ");
		if (testc8.testJump_NNN())
			OutputDebugString("passed");
		else
			OutputDebugString("failed");
	}

	{
		OutputDebugString("\n"); OutputDebugString("testCall_NNN(): ");
		if (testc8.testCall_NNN())
			OutputDebugString("passed");
		else
			OutputDebugString("failed");
	}

	{
		OutputDebugString("\n"); OutputDebugString("testSkipNextIf_Vx_Equal_KK(): ");
		if (testc8.testSkipNextIf_Vx_Equal_KK())
			OutputDebugString("passed");
		else
			OutputDebugString("failed");
	}

	{
		OutputDebugString("\n"); OutputDebugString("testSkipNextIf_Vx_NotEqual_KK(): ");
		if (testc8.testSkipNextIf_Vx_NotEqual_KK())
			OutputDebugString("passed");
		else
			OutputDebugString("failed");
	}

	{
		OutputDebugString("\n"); OutputDebugString("testSkipNextIf_Vx_Equal_Vy(): ");
		if (testc8.testSkipNextIf_Vx_Equal_Vy())
			OutputDebugString("passed");
		else
			OutputDebugString("failed");
	}

	{
		OutputDebugString("\n"); OutputDebugString("testVx_Assign_KK(): ");
		if (testc8.testVx_Assign_KK())
			OutputDebugString("passed");
		else
			OutputDebugString("failed");
	}

	{
		OutputDebugString("\n"); OutputDebugString("testVx_Assign_Vx_Plus_KK(): ");
		if (testc8.testVx_Assign_Vx_Plus_KK())
			OutputDebugString("passed");
		else
			OutputDebugString("failed");
	}

	{
		OutputDebugString("\n"); OutputDebugString("testVx_Assign_Vy(): ");
		if (testc8.testVx_Assign_Vy())
			OutputDebugString("passed");
		else
			OutputDebugString("failed");
	}

	{
		OutputDebugString("\n"); OutputDebugString("testVx_Assign_Vx_OR_Vy(): ");
		if (testc8.testVx_Assign_Vx_OR_Vy())
			OutputDebugString("passed");
		else
			OutputDebugString("failed");
	}

	{
		OutputDebugString("\n"); OutputDebugString("testVx_Assign_Vx_AND_Vy(): ");
		if (testc8.testVx_Assign_Vx_AND_Vy())
			OutputDebugString("passed");
		else
			OutputDebugString("failed");
	}

	{
		OutputDebugString("\n"); OutputDebugString("testVx_Assign_Vx_Plus_Vy(): ");
		if (testc8.testVx_Assign_Vx_Plus_Vy())
			OutputDebugString("passed");
		else
			OutputDebugString("failed");
	}

	{
		OutputDebugString("\n"); OutputDebugString("testVx_Assign_Vx_Minus_Vy(): ");
		if (testc8.testVx_Assign_Vx_Minus_Vy())
			OutputDebugString("passed");
		else
			OutputDebugString("failed");
	}

	{
		OutputDebugString("\n"); OutputDebugString("testVx_Assign_VxVy_ShiftRightBy1(): ");
		if (testc8.testVx_Assign_VxVy_ShiftRightBy1())
			OutputDebugString("passed");
		else
			OutputDebugString("failed");
	}

	{
		OutputDebugString("\n"); OutputDebugString("testVx_Assign_Vy_Minus_Vx(): ");
		if (testc8.testVx_Assign_Vy_Minus_Vx())
			OutputDebugString("passed");
		else
			OutputDebugString("failed");
	}

	{
		OutputDebugString("\n"); OutputDebugString("testVx_Assign_VxVy_ShiftLeftBy1(): ");
		if (testc8.testVx_Assign_VxVy_ShiftLeftBy1())
			OutputDebugString("passed");
		else
			OutputDebugString("failed");
	}

	{
		OutputDebugString("\n"); OutputDebugString("testI_Assign_NNN(): ");
		if (testc8.testI_Assign_NNN())
			OutputDebugString("passed");
		else
			OutputDebugString("failed");
	}

	{
		OutputDebugString("\n"); OutputDebugString("testVx_Assign_Random_AND_KK(): ");
		if (testc8.testVx_Assign_Random_AND_KK())
			OutputDebugString("passed");
		else
			OutputDebugString("failed");
	}

	{
		OutputDebugString("\n"); OutputDebugString("testDrawSprite(): ");
		if (testc8.testDrawSprite())
			OutputDebugString("passed");
		else
			OutputDebugString("failed");
	}

	{
		OutputDebugString("\n"); OutputDebugString("testI_Assign_I_Plus_Vx(): ");
		if (testc8.testI_Assign_I_Plus_Vx())
			OutputDebugString("passed");
		else
			OutputDebugString("failed");
	}

	{
		//TODO: verify if we are really doing the correct thing on this operation, in terms of the borrow bit
		OutputDebugString("\n"); OutputDebugString("testReadV0_To_Vx_FromMemory(): ");
		if (testc8.testReadV0_To_Vx_FromMemory())
			OutputDebugString("passed");
		else
			OutputDebugString("failed");
	}

	OutputDebugString("\n#################\nSuperChip tests\n");

	{
		OutputDebugString("\n"); OutputDebugString("testScrollDown_N_Lines(): ");
		if (testc8.testScrollDown_N_Lines())
			OutputDebugString("passed");
		else
			OutputDebugString("failed");
	}

	{
		OutputDebugString("\n"); OutputDebugString("testScroll_Right(): ");
		if (testc8.testScroll_Right())
			OutputDebugString("passed");
		else
			OutputDebugString("failed");
	}

	{
		OutputDebugString("\n"); OutputDebugString("testScroll_Left(): ");
		if (testc8.testScroll_Left())
			OutputDebugString("passed");
		else
			OutputDebugString("failed");
	}

	{
		OutputDebugString("\n"); OutputDebugString("testExit(): ");
		if (testc8.testExit())
			OutputDebugString("passed");
		else
			OutputDebugString("failed");
	}

	{
		OutputDebugString("\n"); OutputDebugString("testDisableSuperChipMode(): ");
		if (testc8.testDisableSuperChipMode())
			OutputDebugString("passed");
		else
			OutputDebugString("failed");
	}

	{
		OutputDebugString("\n"); OutputDebugString("testEnableSuperChipMode(): ");
		if (testc8.testEnableSuperChipMode())
			OutputDebugString("passed");
		else
			OutputDebugString("failed");
	}

	{
		OutputDebugString("\n"); OutputDebugString("testStoreV0_To_Vx_InHP48(): ");
		if (testc8.testStoreV0_To_Vx_InHP48())
			OutputDebugString("passed");
		else
			OutputDebugString("failed");
	}

	{
		OutputDebugString("\n"); OutputDebugString("testReadV0_To_Vx_FromHP48(): ");
		if (testc8.testReadV0_To_Vx_FromHP48())
			OutputDebugString("passed");
		else
			OutputDebugString("failed");
	}

	OutputDebugString("\n");
	OutputDebugString("\n");

}

void loadTestFile()
{
	//temp: read a test rom and extract it to text
	std::string path = getExecutablePath() + "/resources/Chip8 emulator Logo [Garstyciuks].ch8";
	std::string outPath = getExecutablePath() + "/resources/textout.txt";

	std::ifstream fin(path, std::ifstream::binary);
	std::ofstream fout(outPath, std::fstream::out);

	if (fin.is_open())
	{

		//char c = fin.get() & 0xFF;

		fin.seekg(0, std::ios::end);
		std::size_t filesize = fin.tellg();
		fin.seekg(0, std::ios::beg);
		//char* content = new char[filesize];
		unsigned char memfile[4000];
		memset(memfile, 0, 4000);
		fin.read((char*)memfile, filesize);
		memfile[filesize] = '\0';
		/*OutputDebugString("\n");
		OutputDebugString(&memfile[0]);
		OutputDebugString("hejjj");*/
		/*char newfile[500];
		memset(newfile, 0, 500);
		sprintf(newfile, "%#2x", (unsigned char)memfile[1]);*/
		fout << "{\n";
		for (int i = 0; i < filesize; ++i)
		{
			fout << "0x" << std::hex << std::setprecision(4) << (int)memfile[i] << ", ";
			if (i % 15 == 1)
				fout << std::endl;
		}
		//fout.write((char*)memfile, filesize);

		//int i = 0;
		//while (fin.good())
		//{
		//	++i;
		//	/*char msg[200];
		//	sprintf(msg, "\nC: %x", c);
		//	OutputDebugString(msg);*/
		//	//char newc[2];
		//	//sprintf(newc, "%x", c);
		//	if (i % 2 == 1)
		//		fout << std::hex << std::setprecision(2) << (int)c  << std::endl;
		//	else
		//		fout << std::hex << std::setprecision(2) << (int)c;
		//	c = fin.get() & 0xFF;
		//}

	}
}